/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

/* This file was auto-generated by ensureUiComponents() in @jupyterlab/buildutils */

import { LabIcon } from './labicon';

// icon svg import statements
import addAboveSvgstr from '../../style/icons/toolbar/add-above.svg';
import addBelowSvgstr from '../../style/icons/toolbar/add-below.svg';
import addSvgstr from '../../style/icons/toolbar/add.svg';
import bellSvgstr from '../../style/icons/statusbar/bell.svg';
import bugDotSvgstr from '../../style/icons/toolbar/bug-dot.svg';
import bugSvgstr from '../../style/icons/toolbar/bug.svg';
import buildSvgstr from '../../style/icons/sidebar/build.svg';
import caretDownEmptySvgstr from '../../style/icons/arrow/caret-down-empty.svg';
import caretDownEmptyThinSvgstr from '../../style/icons/arrow/caret-down-empty-thin.svg';
import caretDownSvgstr from '../../style/icons/arrow/caret-down.svg';
import caretLeftSvgstr from '../../style/icons/arrow/caret-left.svg';
import caretRightSvgstr from '../../style/icons/arrow/caret-right.svg';
import caretUpEmptyThinSvgstr from '../../style/icons/arrow/caret-up-empty-thin.svg';
import caretUpSvgstr from '../../style/icons/arrow/caret-up.svg';
import caseSensitiveSvgstr from '../../style/icons/search/case-sensitive.svg';
import checkSvgstr from '../../style/icons/toolbar/check.svg';
import circleEmptySvgstr from '../../style/icons/toolbar/circle-empty.svg';
import circleSvgstr from '../../style/icons/toolbar/circle.svg';
import clearSvgstr from '../../style/icons/toolbar/clear.svg';
import closeSvgstr from '../../style/icons/toolbar/close.svg';
import codeCheckSvgstr from '../../style/icons/lsp/code-check.svg';
import codeSvgstr from '../../style/icons/toolbar/code.svg';
import collapseAllSvgstr from '../../style/icons/toolbar/collapse-all.svg';
import consoleSvgstr from '../../style/icons/filetype/console.svg';
import copySvgstr from '../../style/icons/toolbar/copy.svg';
import copyrightSvgstr from '../../style/icons/licenses/copyright.svg';
import cutSvgstr from '../../style/icons/toolbar/cut.svg';
import deleteSvgstr from '../../style/icons/toolbar/delete.svg';
import downloadSvgstr from '../../style/icons/toolbar/download.svg';
import duplicateSvgstr from '../../style/icons/toolbar/duplicate.svg';
import editSvgstr from '../../style/icons/toolbar/edit.svg';
import ellipsesSvgstr from '../../style/icons/toolbar/ellipses.svg';
import errorSvgstr from '../../style/icons/error.svg';
import expandAllSvgstr from '../../style/icons/toolbar/expand-all.svg';
import extensionSvgstr from '../../style/icons/sidebar/extension.svg';
import fastForwardSvgstr from '../../style/icons/toolbar/fast-forward.svg';
import fileSvgstr from '../../style/icons/filetype/file.svg';
import fileUploadSvgstr from '../../style/icons/toolbar/file-upload.svg';
import filterDotSvgstr from '../../style/icons/search/filter-dot.svg';
import filterListSvgstr from '../../style/icons/toolbar/filter-list.svg';
import filterSvgstr from '../../style/icons/search/filter.svg';
import folderFavoriteSvgstr from '../../style/icons/filetype/folder-favorite.svg';
import folderSvgstr from '../../style/icons/filetype/folder.svg';
import homeSvgstr from '../../style/icons/filetype/home.svg';
import html5Svgstr from '../../style/icons/filetype/html5.svg';
import imageSvgstr from '../../style/icons/filetype/image.svg';
import infoSvgstr from '../../style/icons/info.svg';
import inspectorSvgstr from '../../style/icons/filetype/inspector.svg';
import jsonSvgstr from '../../style/icons/filetype/json.svg';
import juliaSvgstr from '../../style/icons/filetype/julia.svg';
import jupyterFaviconSvgstr from '../../style/icons/jupyter/jupyter-favicon.svg';
import jupyterSvgstr from '../../style/icons/jupyter/jupyter.svg';
import jupyterlabWordmarkSvgstr from '../../style/icons/jupyter/jupyterlab-wordmark.svg';
import kernelSvgstr from '../../style/icons/statusbar/kernel.svg';
import keyboardSvgstr from '../../style/icons/filetype/keyboard.svg';
import launchSvgstr from '../../style/icons/toolbar/launch.svg';
import launcherSvgstr from '../../style/icons/filetype/launcher.svg';
import lineFormSvgstr from '../../style/icons/statusbar/line-form.svg';
import linkSvgstr from '../../style/icons/toolbar/link.svg';
import listSvgstr from '../../style/icons/statusbar/list.svg';
import markdownSvgstr from '../../style/icons/filetype/markdown.svg';
import moveDownSvgstr from '../../style/icons/toolbar/move-down.svg';
import moveUpSvgstr from '../../style/icons/toolbar/move-up.svg';
import newFolderSvgstr from '../../style/icons/toolbar/new-folder.svg';
import notTrustedSvgstr from '../../style/icons/statusbar/not-trusted.svg';
import notebookSvgstr from '../../style/icons/filetype/notebook.svg';
import numberingSvgstr from '../../style/icons/toolbar/numbering.svg';
import offlineBoltSvgstr from '../../style/icons/toolbar/offline-bolt.svg';
import paletteSvgstr from '../../style/icons/sidebar/palette.svg';
import pasteSvgstr from '../../style/icons/toolbar/paste.svg';
import pdfSvgstr from '../../style/icons/filetype/pdf.svg';
import pythonSvgstr from '../../style/icons/filetype/python.svg';
import rKernelSvgstr from '../../style/icons/filetype/r-kernel.svg';
import reactSvgstr from '../../style/icons/filetype/react.svg';
import redoSvgstr from '../../style/icons/toolbar/redo.svg';
import refreshSvgstr from '../../style/icons/toolbar/refresh.svg';
import regexSvgstr from '../../style/icons/search/regex.svg';
import runSvgstr from '../../style/icons/toolbar/run.svg';
import runningSvgstr from '../../style/icons/sidebar/running.svg';
import saveSvgstr from '../../style/icons/toolbar/save.svg';
import searchSvgstr from '../../style/icons/toolbar/search.svg';
import settingsSvgstr from '../../style/icons/filetype/settings.svg';
import shareSvgstr from '../../style/icons/sidebar/share.svg';
import spreadsheetSvgstr from '../../style/icons/filetype/spreadsheet.svg';
import stopSvgstr from '../../style/icons/toolbar/stop.svg';
import tabSvgstr from '../../style/icons/sidebar/tab.svg';
import tableRowsSvgstr from '../../style/icons/toolbar/table-rows.svg';
import tagSvgstr from '../../style/icons/toolbar/tag.svg';
import terminalSvgstr from '../../style/icons/statusbar/terminal.svg';
import textEditorSvgstr from '../../style/icons/filetype/text-editor.svg';
import tocSvgstr from '../../style/icons/sidebar/toc.svg';
import treeViewSvgstr from '../../style/icons/toolbar/tree-view.svg';
import trustedSvgstr from '../../style/icons/statusbar/trusted.svg';
import undoSvgstr from '../../style/icons/toolbar/undo.svg';
import userSvgstr from '../../style/icons/sidebar/user.svg';
import usersSvgstr from '../../style/icons/sidebar/users.svg';
import vegaSvgstr from '../../style/icons/filetype/vega.svg';
import wordSvgstr from '../../style/icons/search/word.svg';
import yamlSvgstr from '../../style/icons/filetype/yaml.svg';

// LabIcon instance construction
export const addAboveIcon = new LabIcon({ name: 'ui-components:add-above', svgstr: addAboveSvgstr });
export const addBelowIcon = new LabIcon({ name: 'ui-components:add-below', svgstr: addBelowSvgstr });
export const addIcon = new LabIcon({ name: 'ui-components:add', svgstr: addSvgstr });
export const bellIcon = new LabIcon({ name: 'ui-components:bell', svgstr: bellSvgstr });
export const bugDotIcon = new LabIcon({ name: 'ui-components:bug-dot', svgstr: bugDotSvgstr });
export const bugIcon = new LabIcon({ name: 'ui-components:bug', svgstr: bugSvgstr });
export const buildIcon = new LabIcon({ name: 'ui-components:build', svgstr: buildSvgstr });
export const caretDownEmptyIcon = new LabIcon({ name: 'ui-components:caret-down-empty', svgstr: caretDownEmptySvgstr });
export const caretDownEmptyThinIcon = new LabIcon({ name: 'ui-components:caret-down-empty-thin', svgstr: caretDownEmptyThinSvgstr });
export const caretDownIcon = new LabIcon({ name: 'ui-components:caret-down', svgstr: caretDownSvgstr });
export const caretLeftIcon = new LabIcon({ name: 'ui-components:caret-left', svgstr: caretLeftSvgstr });
export const caretRightIcon = new LabIcon({ name: 'ui-components:caret-right', svgstr: caretRightSvgstr });
export const caretUpEmptyThinIcon = new LabIcon({ name: 'ui-components:caret-up-empty-thin', svgstr: caretUpEmptyThinSvgstr });
export const caretUpIcon = new LabIcon({ name: 'ui-components:caret-up', svgstr: caretUpSvgstr });
export const caseSensitiveIcon = new LabIcon({ name: 'ui-components:case-sensitive', svgstr: caseSensitiveSvgstr });
export const checkIcon = new LabIcon({ name: 'ui-components:check', svgstr: checkSvgstr });
export const circleEmptyIcon = new LabIcon({ name: 'ui-components:circle-empty', svgstr: circleEmptySvgstr });
export const circleIcon = new LabIcon({ name: 'ui-components:circle', svgstr: circleSvgstr });
export const clearIcon = new LabIcon({ name: 'ui-components:clear', svgstr: clearSvgstr });
export const closeIcon = new LabIcon({ name: 'ui-components:close', svgstr: closeSvgstr });
export const codeCheckIcon = new LabIcon({ name: 'ui-components:code-check', svgstr: codeCheckSvgstr });
export const codeIcon = new LabIcon({ name: 'ui-components:code', svgstr: codeSvgstr });
export const collapseAllIcon = new LabIcon({ name: 'ui-components:collapse-all', svgstr: collapseAllSvgstr });
export const consoleIcon = new LabIcon({ name: 'ui-components:console', svgstr: consoleSvgstr });
export const copyIcon = new LabIcon({ name: 'ui-components:copy', svgstr: copySvgstr });
export const copyrightIcon = new LabIcon({ name: 'ui-components:copyright', svgstr: copyrightSvgstr });
export const cutIcon = new LabIcon({ name: 'ui-components:cut', svgstr: cutSvgstr });
export const deleteIcon = new LabIcon({ name: 'ui-components:delete', svgstr: deleteSvgstr });
export const downloadIcon = new LabIcon({ name: 'ui-components:download', svgstr: downloadSvgstr });
export const duplicateIcon = new LabIcon({ name: 'ui-components:duplicate', svgstr: duplicateSvgstr });
export const editIcon = new LabIcon({ name: 'ui-components:edit', svgstr: editSvgstr });
export const ellipsesIcon = new LabIcon({ name: 'ui-components:ellipses', svgstr: ellipsesSvgstr });
export const errorIcon = new LabIcon({ name: 'ui-components:error', svgstr: errorSvgstr });
export const expandAllIcon = new LabIcon({ name: 'ui-components:expand-all', svgstr: expandAllSvgstr });
export const extensionIcon = new LabIcon({ name: 'ui-components:extension', svgstr: extensionSvgstr });
export const fastForwardIcon = new LabIcon({ name: 'ui-components:fast-forward', svgstr: fastForwardSvgstr });
export const fileIcon = new LabIcon({ name: 'ui-components:file', svgstr: fileSvgstr });
export const fileUploadIcon = new LabIcon({ name: 'ui-components:file-upload', svgstr: fileUploadSvgstr });
export const filterDotIcon = new LabIcon({ name: 'ui-components:filter-dot', svgstr: filterDotSvgstr });
export const filterIcon = new LabIcon({ name: 'ui-components:filter', svgstr: filterSvgstr });
export const filterListIcon = new LabIcon({ name: 'ui-components:filter-list', svgstr: filterListSvgstr });
export const folderFavoriteIcon = new LabIcon({ name: 'ui-components:folder-favorite', svgstr: folderFavoriteSvgstr });
export const folderIcon = new LabIcon({ name: 'ui-components:folder', svgstr: folderSvgstr });
export const homeIcon = new LabIcon({ name: 'ui-components:home', svgstr: homeSvgstr });
export const html5Icon = new LabIcon({ name: 'ui-components:html5', svgstr: html5Svgstr });
export const imageIcon = new LabIcon({ name: 'ui-components:image', svgstr: imageSvgstr });
export const infoIcon = new LabIcon({ name: 'ui-components:info', svgstr: infoSvgstr });
export const inspectorIcon = new LabIcon({ name: 'ui-components:inspector', svgstr: inspectorSvgstr });
export const jsonIcon = new LabIcon({ name: 'ui-components:json', svgstr: jsonSvgstr });
export const juliaIcon = new LabIcon({ name: 'ui-components:julia', svgstr: juliaSvgstr });
export const jupyterFaviconIcon = new LabIcon({ name: 'ui-components:jupyter-favicon', svgstr: jupyterFaviconSvgstr });
export const jupyterIcon = new LabIcon({ name: 'ui-components:jupyter', svgstr: jupyterSvgstr });
export const jupyterlabWordmarkIcon = new LabIcon({ name: 'ui-components:jupyterlab-wordmark', svgstr: jupyterlabWordmarkSvgstr });
export const kernelIcon = new LabIcon({ name: 'ui-components:kernel', svgstr: kernelSvgstr });
export const keyboardIcon = new LabIcon({ name: 'ui-components:keyboard', svgstr: keyboardSvgstr });
export const launchIcon = new LabIcon({ name: 'ui-components:launch', svgstr: launchSvgstr });
export const launcherIcon = new LabIcon({ name: 'ui-components:launcher', svgstr: launcherSvgstr });
export const lineFormIcon = new LabIcon({ name: 'ui-components:line-form', svgstr: lineFormSvgstr });
export const linkIcon = new LabIcon({ name: 'ui-components:link', svgstr: linkSvgstr });
export const listIcon = new LabIcon({ name: 'ui-components:list', svgstr: listSvgstr });
export const markdownIcon = new LabIcon({ name: 'ui-components:markdown', svgstr: markdownSvgstr });
export const moveDownIcon = new LabIcon({ name: 'ui-components:move-down', svgstr: moveDownSvgstr });
export const moveUpIcon = new LabIcon({ name: 'ui-components:move-up', svgstr: moveUpSvgstr });
export const newFolderIcon = new LabIcon({ name: 'ui-components:new-folder', svgstr: newFolderSvgstr });
export const notTrustedIcon = new LabIcon({ name: 'ui-components:not-trusted', svgstr: notTrustedSvgstr });
export const notebookIcon = new LabIcon({ name: 'ui-components:notebook', svgstr: notebookSvgstr });
export const numberingIcon = new LabIcon({ name: 'ui-components:numbering', svgstr: numberingSvgstr });
export const offlineBoltIcon = new LabIcon({ name: 'ui-components:offline-bolt', svgstr: offlineBoltSvgstr });
export const paletteIcon = new LabIcon({ name: 'ui-components:palette', svgstr: paletteSvgstr });
export const pasteIcon = new LabIcon({ name: 'ui-components:paste', svgstr: pasteSvgstr });
export const pdfIcon = new LabIcon({ name: 'ui-components:pdf', svgstr: pdfSvgstr });
export const pythonIcon = new LabIcon({ name: 'ui-components:python', svgstr: pythonSvgstr });
export const rKernelIcon = new LabIcon({ name: 'ui-components:r-kernel', svgstr: rKernelSvgstr });
export const reactIcon = new LabIcon({ name: 'ui-components:react', svgstr: reactSvgstr });
export const redoIcon = new LabIcon({ name: 'ui-components:redo', svgstr: redoSvgstr });
export const refreshIcon = new LabIcon({ name: 'ui-components:refresh', svgstr: refreshSvgstr });
export const regexIcon = new LabIcon({ name: 'ui-components:regex', svgstr: regexSvgstr });
export const runIcon = new LabIcon({ name: 'ui-components:run', svgstr: runSvgstr });
export const runningIcon = new LabIcon({ name: 'ui-components:running', svgstr: runningSvgstr });
export const saveIcon = new LabIcon({ name: 'ui-components:save', svgstr: saveSvgstr });
export const searchIcon = new LabIcon({ name: 'ui-components:search', svgstr: searchSvgstr });
export const settingsIcon = new LabIcon({ name: 'ui-components:settings', svgstr: settingsSvgstr });
export const shareIcon = new LabIcon({ name: 'ui-components:share', svgstr: shareSvgstr });
export const spreadsheetIcon = new LabIcon({ name: 'ui-components:spreadsheet', svgstr: spreadsheetSvgstr });
export const stopIcon = new LabIcon({ name: 'ui-components:stop', svgstr: stopSvgstr });
export const tabIcon = new LabIcon({ name: 'ui-components:tab', svgstr: tabSvgstr });
export const tableRowsIcon = new LabIcon({ name: 'ui-components:table-rows', svgstr: tableRowsSvgstr });
export const tagIcon = new LabIcon({ name: 'ui-components:tag', svgstr: tagSvgstr });
export const terminalIcon = new LabIcon({ name: 'ui-components:terminal', svgstr: terminalSvgstr });
export const textEditorIcon = new LabIcon({ name: 'ui-components:text-editor', svgstr: textEditorSvgstr });
export const tocIcon = new LabIcon({ name: 'ui-components:toc', svgstr: tocSvgstr });
export const treeViewIcon = new LabIcon({ name: 'ui-components:tree-view', svgstr: treeViewSvgstr });
export const trustedIcon = new LabIcon({ name: 'ui-components:trusted', svgstr: trustedSvgstr });
export const undoIcon = new LabIcon({ name: 'ui-components:undo', svgstr: undoSvgstr });
export const userIcon = new LabIcon({ name: 'ui-components:user', svgstr: userSvgstr });
export const usersIcon = new LabIcon({ name: 'ui-components:users', svgstr: usersSvgstr });
export const vegaIcon = new LabIcon({ name: 'ui-components:vega', svgstr: vegaSvgstr });
export const wordIcon = new LabIcon({ name: 'ui-components:word', svgstr: wordSvgstr });
export const yamlIcon = new LabIcon({ name: 'ui-components:yaml', svgstr: yamlSvgstr });
