/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

/* TODO: finish cleaning up these cases of icon-as-css-background-image */
:root {
  /* may be unused */

  /*--jp-icon-bug: url('icons/md/bug.svg');*/

  /* blocked by lumino interaction */

  /*--jp-icon-caret-down: url('icons/md/caretdown.svg');*/

  /*--jp-icon-caret-right: url('icons/md/caretright.svg');*/

  /*--jp-icon-caret-up: url('icons/md/caretup.svg');*/

  /*--jp-icon-caret-left: url('icons/md/caretleft.svg');*/

  /*--jp-icon-search: url('icons/md/search.svg');*/

  /*--jp-icon-search-white: url('icons/md/search-white.svg');*/
}
