/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

const defaultConfig = require('@jupyterlab/testing/lib/babel-config');

module.exports = defaultConfig;
