/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Distributed under the terms of the Modified BSD License.
|----------------------------------------------------------------------------*/

/* This file was auto-generated by ensurePackage() in @jupyterlab/buildutils */
@import url('~@lumino/widgets/style/index.css');
@import url('~@jupyterlab/ui-components/style/index.css');
@import url('~@jupyterlab/statusbar/style/index.css');
@import url('~@jupyterlab/apputils/style/index.css');
@import url('~@jupyterlab/codeeditor/style/index.css');
@import url('~@jupyterlab/rendermime/style/index.css');
@import url('~@jupyterlab/docregistry/style/index.css');
@import url('~@jupyterlab/application/style/index.css');
@import url('~@jupyterlab/documentsearch/style/index.css');
@import url('~@jupyterlab/codemirror/style/index.css');
@import url('~@jupyterlab/docmanager/style/index.css');
@import url('~@jupyterlab/filebrowser/style/index.css');
@import url('~@jupyterlab/toc/style/index.css');
@import url('~@jupyterlab/cells/style/index.css');
@import url('~@jupyterlab/completer/style/index.css');
@import url('~@jupyterlab/docmanager-extension/style/index.css');
@import url('~@jupyterlab/launcher/style/index.css');
@import url('~@jupyterlab/logconsole/style/index.css');
@import url('~@jupyterlab/lsp/style/index.css');
@import url('~@jupyterlab/mainmenu/style/index.css');
@import url('~@jupyterlab/notebook/style/index.css');
@import url('~@jupyterlab/metadataform/style/index.css');
@import url('~@jupyterlab/property-inspector/style/index.css');
