/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

import * as config from '@jupyterlab/testing/lib/babel-config';

module.exports = config;
